import enum

from zigpy.types import uint8_t


class NvItemsIds(uint8_t, enum.Enum):
    EXTADDR = 1
    BOOTCOUNTER = 2
    STARTUP_OPTION = 3
    START_DELAY = 4
    NIB = 33
    DEVICE_LIST = 34
    ADDRMGR = 35
    POLL_RATE = 36
    QUEUED_POLL_RATE = 37
    RESPONSE_POLL_RATE = 38
    REJOIN_POLL_RATE = 39
    DATA_RETRIES = 40
    POLL_FAILURE_RETRIES = 41
    STACK_PROFILE = 42
    INDIRECT_MSG_TIMEOUT = 43
    ROUTE_EXPIRY_TIME = 44
    EXTENDED_PAN_ID = 45
    BCAST_RETRIES = 46
    PASSIVE_ACK_TIMEOUT = 47
    BCAST_DELIVERY_TIME = 48
    NWK_MODE = 49
    CONCENTRATOR_ENABLE = 50
    CONCENTRATOR_DISCOVERY = 51
    CONCENTRATOR_RADIUS = 52
    CONCENTRATOR_RC = 54
    NWK_MGR_MODE = 55
    SRC_RTG_EXPIRY_TIME = 56
    ROUTE_DISCOVERY_TIME = 57
    NWK_ACTIVE_KEY_INFO = 58
    NWK_ALTERN_KEY_INFO = 59
    ROUTER_OFF_ASSOC_CLEANUP = 60
    NWK_LEAVE_REQ_ALLOWED = 61
    NWK_CHILD_AGE_ENABLE = 62
    DEVICE_LIST_KA_TIMEOUT = 63
    BINDING_TABLE = 65
    GROUP_TABLE = 66
    APS_FRAME_RETRIES = 67
    APS_ACK_WAIT_DURATION = 68
    APS_ACK_WAIT_MULTIPLIER = 69
    BINDING_TIME = 70
    APS_USE_EXT_PANID = 71
    APS_USE_INSECURE_JOIN = 72
    COMMISSIONED_NWK_ADDR = 73
    APS_NONMEMBER_RADIUS = 75
    APS_LINK_KEY_TABLE = 76
    APS_DUPREJ_TIMEOUT_INC = 77
    APS_DUPREJ_TIMEOUT_COUNT = 78
    APS_DUPREJ_TABLE_SIZE = 79
    DIAGNOSTIC_STATS = 80
    BDBNODEISONANETWORK = 85
    SECURITY_LEVEL = 97
    PRECFGKEY = 98
    PRECFGKEYS_ENABLE = 99
    SECURITY_MODE = 100
    SECURE_PERMIT_JOIN = 101
    APS_LINK_KEY_TYPE = 102
    APS_ALLOW_R19_SECURITY = 103
    IMPLICIT_CERTIFICATE = 105
    DEVICE_PRIVATE_KEY = 106
    CA_PUBLIC_KEY = 107
    KE_MAX_DEVICES = 108
    USE_DEFAULT_TCLK = 109
    RNG_COUNTER = 111
    RANDOM_SEED = 112
    TRUSTCENTER_ADDR = 113
    LEGACY_NWK_SEC_MATERIAL_TABLE_START = 117  # Valid for <= Z-Stack 3.0.x
    EX_NWK_SEC_MATERIAL_TABLE = 7  # Valid for >= Z-Stack 3.x.0
    USERDESC = 129
    NWKKEY = 130
    PANID = 131
    CHANLIST = 132
    LEAVE_CTRL = 133
    SCAN_DURATION = 134
    LOGICAL_TYPE = 135
    NWKMGR_MIN_TX = 136
    NWKMGR_ADDR = 137
    ZDO_DIRECT_CB = 143
    SCENE_TABLE = 145
    MIN_FREE_NWK_ADDR = 146
    MAX_FREE_NWK_ADDR = 147
    MIN_FREE_GRP_ID = 148
    MAX_FREE_GRP_ID = 149
    MIN_GRP_IDS = 150
    MAX_GRP_IDS = 151
    OTA_BLOCK_REQ_DELAY = 152
    SAPI_ENDPOINT = 161
    SAS_SHORT_ADDR = 177
    SAS_EXT_PANID = 178
    SAS_PANID = 179
    SAS_CHANNEL_MASK = 180
    SAS_PROTOCOL_VER = 181
    SAS_STACK_PROFILE = 182
    SAS_STARTUP_CTRL = 183
    SAS_TC_ADDR = 193
    SAS_TC_MASTER_KEY = 194
    SAS_NWK_KEY = 195
    SAS_USE_INSEC_JOIN = 196
    SAS_PRECFG_LINK_KEY = 197
    SAS_NWK_KEY_SEQ_NUM = 198
    SAS_NWK_KEY_TYPE = 199
    SAS_NWK_MGR_ADDR = 200
    SAS_CURR_TC_MASTER_KEY = 209
    SAS_CURR_NWK_KEY = 210
    SAS_CURR_PRECFG_LINK_KEY = 211
    LEGACY_TCLK_TABLE_START = 257  # Valid for <= Z-Stack 3.0.x
    LEGACY_TCLK_TABLE_END = 511  # Valid for <= Z-Stack 3.0.x
    EX_TCLK_TABLE = 4  # Valid for >= Z-Stack 3.0.x
    APS_LINK_KEY_DATA_START = 513
    APS_LINK_KEY_DATA_END = 767
    DUPLICATE_BINDING_TABLE = 768
    DUPLICATE_DEVICE_LIST = 769
    DUPLICATE_DEVICE_LIST_KA_TIMEOUT = 770
    ZNP_HAS_CONFIGURED_ZSTACK1 = 3840
    ZNP_HAS_CONFIGURED_ZSTACK3 = 96


class Common:
    devStates = {
        "HOLD": 0,
        "INIT": 1,
        "NWK_DISC": 2,
        "NWK_JOINING": 3,
        "NWK_REJOIN": 4,
        "END_DEVICE_UNAUTH": 5,
        "END_DEVICE": 6,
        "ROUTER": 7,
        "COORD_STARTING": 8,
        "ZB_COORD": 9,
        "NWK_ORPHAN": 10,
        "INVALID_REQTYPE": 128,
        "DEVICE_NOT_FOUND": 129,
        "INVALID_EP": 130,
        "NOT_ACTIVE": 131,
        "NOT_SUPPORTED": 132,
        "TIMEOUT": 133,
        "NO_MATCH": 134,
        "NO_ENTRY": 136,
        "NO_DESCRIPTOR": 137,
        "INSUFFICIENT_SPACE": 138,
        "NOT_PERMITTED": 139,
        "TABLE_FULL": 140,
        "NOT_AUTHORIZED": 141,
        "BINDING_TABLE_FULL": 142,
    }
    logicalChannels = {
        "NONE": 0,
        "CH11": 11,
        "CH12": 12,
        "CH13": 13,
        "CH14": 14,
        "CH15": 15,
        "CH16": 16,
        "CH17": 17,
        "CH18": 18,
        "CH19": 19,
        "CH20": 20,
        "CH21": 21,
        "CH22": 22,
        "CH23": 23,
        "CH24": 24,
        "CH25": 25,
        "CH26": 26,
    }
    channelMask = {
        "CH11": 2048,
        "CH12": 4096,
        "CH13": 8192,
        "CH14": 16384,
        "CH15": 32768,
        "CH16": 65536,
        "CH17": 131072,
        "CH18": 262144,
        "CH19": 524288,
        "CH20": 1048576,
        "CH21": 2097152,
        "CH22": 4194304,
        "CH23": 8388608,
        "CH24": 16777216,
        "CH25": 33554432,
        "CH26": 67108864,
        "CH_ALL": 134215680,
    }



"""Setup module for zigpy-cc"""

import os

from setuptools import find_packages, setup

import zigpy_cc

this_directory = os.path.join(os.path.abspath(os.path.dirname(__file__)))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="zigpy-cc",
    version=zigpy_cc.__version__,
    description="A library which communicates with "
    "Texas Instruments CC2531 radios for zigpy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/zigpy/zigpy-cc",
    author="Balazs Sandor",
    author_email="sanyatuning@gmail.com",
    license="GPL-3.0",
    packages=find_packages(exclude=["*.tests"]),
    install_requires=["pyserial-asyncio", "zigpy>=0.20.a1"],
    tests_require=["asynctest", "pytest", "pytest-asyncio"],
)



import asyncio
import logging
import os

from zigpy.zcl.clusters.general import Ota
from zigpy.zcl.clusters.security import IasZone, IasWd

from zigpy_cc.api import API
from zigpy_cc.const import Constants
from zigpy_cc.types import NetworkOptions, Subsystem, ZnpVersion, CommandType
from zigpy_cc.exception import CommandError
from zigpy_cc.zigbee.backup import Restore
from zigpy_cc.zigbee.common import Common
from .nv_items import Items

LOGGER = logging.getLogger(__name__)


class Endpoint:
    def __init__(self, **kwargs) -> None:
        self.endpoint = None
        self.appdeviceid = 0x0005
        self.appdevver = 0
        self.appnuminclusters = 0
        self.appinclusterlist = []
        self.appnumoutclusters = 0
        self.appoutclusterlist = []
        self.latencyreq = Constants.AF.networkLatencyReq.NO_LATENCY_REQS
        for key, value in kwargs.items():
            setattr(self, key, value)


Endpoints = [
    Endpoint(endpoint=1, appprofid=0x0104),
    Endpoint(endpoint=2, appprofid=0x0101),
    Endpoint(endpoint=3, appprofid=0x0106),
    Endpoint(endpoint=4, appprofid=0x0107),
    Endpoint(endpoint=5, appprofid=0x0108),
    Endpoint(endpoint=6, appprofid=0x0109),
    Endpoint(endpoint=8, appprofid=0x0104),
    Endpoint(
        endpoint=11,
        appprofid=0x0104,
        appdeviceid=0x0400,
        appnumoutclusters=2,
        appoutclusterlist=[IasZone.cluster_id, IasWd.cluster_id],
    ),
    # TERNCY: https://github.com/Koenkk/zigbee-herdsman/issues/82
    Endpoint(endpoint=0x6E, appprofid=0x0104),
    Endpoint(endpoint=12, appprofid=0xC05E),
    Endpoint(
        endpoint=13,
        appprofid=0x0104,
        appnuminclusters=1,
        appinclusterlist=[Ota.cluster_id],
    ),
    # Insta/Jung/Gira: OTA fallback EP (since it's buggy in firmware 10023202
    # when it tries to find a matching EP for OTA - it queries for ZLL profile,
    # but then contacts with HA profile)
    Endpoint(endpoint=47, appprofid=0x0104),
    Endpoint(endpoint=242, appprofid=0xA1E0),
]


async def validate_item(
    znp: API,
    item,
    message,
    subsystem=Subsystem.SYS,
    command="osalNvRead",
    expected_status=None,
):
    result = await znp.request(subsystem, command, item, None, expected_status)
    if result.payload["value"] != item["value"]:
        msg = "Item '{}' is invalid, got '{}', expected '{}'".format(
            message, result.payload["value"], item["value"]
        )
        LOGGER.debug(msg)
        raise AssertionError(msg)
    else:
        LOGGER.debug("Item '%s' is valid", message)


async def needsToBeInitialised(znp: API, version, options):
    try:
        await validate_item(
            znp,
            Items.znpHasConfigured(version),
            "hasConfigured",
            expected_status=[0, 2],
        )
        await validate_item(znp, Items.channelList(options.channelList), "channelList")
        await validate_item(
            znp,
            Items.networkKeyDistribute(options.networkKeyDistribute),
            "networkKeyDistribute",
        )

        if version == ZnpVersion.zStack3x0:
            await validate_item(znp, Items.networkKey(options.networkKey), "networkKey")
        else:
            await validate_item(
                znp,
                Items.networkKey(options.networkKey),
                "networkKey",
                Subsystem.SAPI,
                "readConfiguration",
            )

        try:
            await validate_item(znp, Items.panID(options.panID), "panID")
            await validate_item(
                znp, Items.extendedPanID(options.extendedPanID), "extendedPanID"
            )
        except AssertionError as e:
            if version == ZnpVersion.zStack30x or version == ZnpVersion.zStack3x0:
                # When the panID has never been set, it will be [0xFF, 0xFF].
                result = await znp.request(
                    Subsystem.SYS, "osalNvRead", Items.panID(options.panID)
                )
                LOGGER.debug("PANID: %s", result.payload["value"])
                if result.payload["value"] == bytes([0xFF, 0xFF]):
                    LOGGER.debug("Skip enforcing panID because a random panID is used")
                else:
                    raise e

        return False
    except AssertionError as e:
        LOGGER.debug("Error while validating items: %s", e)
        return True


async def boot(znp: API):
    result = await znp.request(Subsystem.UTIL, "getDeviceInfo", {})

    if result.payload["devicestate"] != Common.devStates["ZB_COORD"]:
        LOGGER.debug("Start ZNP as coordinator...")
        started = znp.wait_for(
            CommandType.AREQ, Subsystem.ZDO, "stateChangeInd", {"state": 9}, 60000
        )
        await znp.request(
            Subsystem.ZDO, "startupFromApp", {"startdelay": 100}, None, [0, 1]
        )
        await started.wait()
        LOGGER.info("ZNP started as coordinator")
    else:
        LOGGER.info("ZNP is already started as coordinator")


async def registerEndpoints(znp: API):
    LOGGER.debug("Register endpoints...")

    active_ep_response = znp.wait_for(CommandType.AREQ, Subsystem.ZDO, "activeEpRsp")
    asyncio.create_task(
        znp.request(
            Subsystem.ZDO, "activeEpReq", {"dstaddr": 0, "nwkaddrofinterest": 0}
        )
    )
    active_ep = await active_ep_response.wait()

    for endpoint in Endpoints:
        if endpoint.endpoint in active_ep.payload["activeeplist"]:
            LOGGER.debug("Endpoint '%s' already registered", endpoint.endpoint)
        else:
            LOGGER.debug("Registering endpoint '%s'", endpoint.endpoint)
            await znp.request(Subsystem.AF, "register", vars(endpoint))


async def initialise(znp: API, version, options: NetworkOptions):
    await znp.request(Subsystem.SYS, "resetReq", {"type": Constants.SYS.resetType.SOFT})
    await znp.request(Subsystem.SYS, "osalNvWrite", Items.startupOption(0x02))
    await znp.request(Subsystem.SYS, "resetReq", {"type": Constants.SYS.resetType.SOFT})
    await znp.request(
        Subsystem.SYS,
        "osalNvWrite",
        Items.logicalType(Constants.ZDO.deviceLogicalType.COORDINATOR),
    )
    await znp.request(
        Subsystem.SYS,
        "osalNvWrite",
        Items.networkKeyDistribute(options.networkKeyDistribute),
    )
    await znp.request(Subsystem.SYS, "osalNvWrite", Items.zdoDirectCb())
    await znp.request(
        Subsystem.SYS, "osalNvWrite", Items.channelList(options.channelList)
    )
    await znp.request(Subsystem.SYS, "osalNvWrite", Items.panID(options.panID))
    await znp.request(
        Subsystem.SYS, "osalNvWrite", Items.extendedPanID(options.extendedPanID)
    )

    if version == ZnpVersion.zStack30x or version == ZnpVersion.zStack3x0:
        await znp.request(
            Subsystem.SYS, "osalNvWrite", Items.networkKey(options.networkKey)
        )

        # Default link key is already OK for Z-Stack 3 ('ZigBeeAlliance09')
        await znp.request(
            Subsystem.APP_CNF,
            "bdbSetChannel",
            {"isPrimary": 0x1, "channel": options.channelList},
        )
        await znp.request(
            Subsystem.APP_CNF, "bdbSetChannel", {"isPrimary": 0x0, "channel": 0x0}
        )

        started = znp.wait_for(
            CommandType.AREQ, Subsystem.ZDO, "stateChangeInd", {"state": 9}, 60000
        )
        await znp.request(Subsystem.APP_CNF, "bdbStartCommissioning", {"mode": 0x04})
        try:
            await started.wait()
        except Exception:
            raise Exception(
                "Coordinator failed to start, probably the panID is already in use, "
                "try a different panID or channel"
            )

        await znp.request(Subsystem.APP_CNF, "bdbStartCommissioning", {"mode": 0x02})
    else:
        await znp.request(
            Subsystem.SAPI, "writeConfiguration", Items.networkKey(options.networkKey)
        )
        await znp.request(Subsystem.SYS, "osalNvWrite", Items.tcLinkKey())

    # expect status code 9 (= item created and initialized)
    await znp.request(
        Subsystem.SYS,
        "osalNvItemInit",
        Items.znpHasConfiguredInit(version),
        None,
        [0, 9],
    )
    await znp.request(Subsystem.SYS, "osalNvWrite", Items.znpHasConfigured(version))


async def addToGroup(znp: API, endpoint: int, group: int):
    result = await znp.request(
        Subsystem.ZDO,
        "extFindGroup",
        {"endpoint": endpoint, "groupid": group},
        None,
        [0, 1],
    )
    if result.payload["status"] == 1:
        await znp.request(
            Subsystem.ZDO,
            "extAddGroup",
            {
                "endpoint": endpoint,
                "groupid": group,
                "namelen": 0,
                "groupname": bytes([]),
            },
        )


async def start_znp(
    znp: API, version, options: NetworkOptions, greenPowerGroup: int, backupPath=""
):
    result = "resumed"

    try:
        await validate_item(
            znp,
            Items.znpHasConfigured(version),
            "hasConfigured",
            expected_status=[0, 2],
        )
        hasConfigured = True
    except (AssertionError, CommandError):
        hasConfigured = False

    if backupPath and os.path.exists(backupPath) and not hasConfigured:
        LOGGER.debug("Restoring coordinator from backup")
        await Restore(znp, backupPath, options)
        result = "restored"
    elif await needsToBeInitialised(znp, version, options):
        LOGGER.debug("Initialize coordinator")
        await initialise(znp, version, options)

        if version == ZnpVersion.zStack12:
            # zStack12 allows to restore a network without restoring a backup
            # (as long as the network key, panID and channel don't change).
            # If the device has not been configured yet we assume that this is the case.
            # If we always return 'reset'
            # the controller clears the database on a reflash of the stick.
            result = "reset" if hasConfigured else "restored"
        else:
            result = "reset"

    await boot(znp)
    await registerEndpoints(znp)

    # Add to required group to receive greenPower messages.
    await addToGroup(znp, 242, greenPowerGroup)

    if result == "restored":
        # Write channel list again, otherwise it doesnt seem to stick.
        await znp.request(
            Subsystem.SYS, "osalNvWrite", Items.channelList(options.channelList)
        )

    return result
